@@ -23,46 +23,42 @@ using tornado.
 
 """
 
-from __future__ import with_statement
+import pickle
+import sys
 import warnings
+from queue import Queue
+from typing import Any, Callable, List, Optional, Sequence, Union, cast, overload
 
 import zmq
+from zmq._typing import Literal
 from zmq.utils import jsonapi
 
-try:
-    import cPickle as pickle
-except ImportError:
-    import pickle
-
 from .ioloop import IOLoop, gen_log
 
 try:
-    from tornado import stack_context
+    import tornado.ioloop
+    from tornado.stack_context import wrap as stack_context_wrap  # type: ignore
 except ImportError:
-    from .minitornado import stack_context
-
-try:
-    from queue import Queue
-except ImportError:
-    from Queue import Queue
-
-from zmq.utils.strtypes import basestring
-
-try:
-    callable
-except NameError:
-    callable = lambda obj: hasattr(obj, '__call__')
+    if "zmq.eventloop.minitornado" in sys.modules:
+        from .minitornado.stack_context import (
+            wrap as stack_context_wrap,  # type: ignore
+        )
+    else:
+        # tornado 5 deprecates stack_context,
+        # tornado 6 removes it
+        def stack_context_wrap(callback):
+            return callback
 
 
-class ZMQStream(object):
+class ZMQStream:
     """A utility class to register callbacks when a zmq socket sends and receives
-    
+
     For use with zmq.eventloop.ioloop
 
     There are three main methods
-    
+
     Methods:
-    
+
     * **on_recv(callback, copy=True):**
         register a callback to be run every time the socket has something to receive
     * **on_send(callback):**
@@ -70,44 +66,46 @@ class ZMQStream(object):
     * **send(self, msg, flags=0, copy=False, callback=None):**
         perform a send that will trigger the callback
         if callback is passed, on_send is also called.
-        
+
         There are also send_multipart(), send_json(), send_pyobj()
-    
+
     Three other methods for deactivating the callbacks:
-    
+
     * **stop_on_recv():**
         turn off the recv callback
     * **stop_on_send():**
         turn off the send callback
-    
+
     which simply call ``on_<evt>(None)``.
-    
+
     The entire socket interface, excluding direct recv methods, is also
     provided, primarily through direct-linking the methods.
     e.g.
-    
+
     >>> stream.bind is stream.socket.bind
     True
-    
+
     """
 
-    socket = None
-    io_loop = None
-    poller = None
-    _send_queue = None
-    _recv_callback = None
-    _send_callback = None
-    _close_callback = None
-    _state = 0
-    _flushed = False
-    _recv_copy = False
-    _fd = None
+    socket: zmq.Socket
+    io_loop: "tornado.ioloop.IOLoop"
+    poller: zmq.Poller
+    _send_queue: Queue
+    _recv_callback: Optional[Callable]
+    _send_callback: Optional[Callable]
+    _close_callback = Optional[Callable]
+    _state: int = 0
+    _flushed: bool = False
+    _recv_copy: bool = False
+    _fd: int
 
-    def __init__(self, socket, io_loop=None):
+    def __init__(
+        self, socket: "zmq.Socket", io_loop: Optional["tornado.ioloop.IOLoop"] = None
+    ):
         self.socket = socket
-        self.io_loop = io_loop or IOLoop.instance()
+        self.io_loop = io_loop or IOLoop.current()
         self.poller = zmq.Poller()
-        self._fd = self.socket.FD
+        self._fd = cast(int, self.socket.FD)
 
         self._send_queue = Queue()
         self._recv_callback = None
@@ -133,33 +131,74 @@ class ZMQStream(object):
     def stop_on_recv(self):
         """Disable callback and automatic receiving."""
         return self.on_recv(None)
-    
+
     def stop_on_send(self):
         """Disable callback on sending."""
         return self.on_send(None)
-    
+
     def stop_on_err(self):
         """DEPRECATED, does nothing"""
         gen_log.warn("on_err does nothing, and will be removed")
-    
-    def on_err(self, callback):
+
+    def on_err(self, callback: Callable):
         """DEPRECATED, does nothing"""
         gen_log.warn("on_err does nothing, and will be removed")
-    
-    def on_recv(self, callback, copy=True):
+
+    @overload
+    def on_recv(
+        self,
+        callback: Callable[[List[bytes]], Any],
+    ) -> None:
+        ...
+
+    @overload
+    def on_recv(
+        self,
+        callback: Callable[[List[bytes]], Any],
+        copy: Literal[True],
+    ) -> None:
+        ...
+
+    @overload
+    def on_recv(
+        self,
+        callback: Callable[[List[zmq.Frame]], Any],
+        copy: Literal[False],
+    ) -> None:
+        ...
+
+    @overload
+    def on_recv(
+        self,
+        callback: Union[
+            Callable[[List[zmq.Frame]], Any],
+            Callable[[List[bytes]], Any],
+        ],
+        copy: bool = ...,
+    ):
+        ...
+
+    def on_recv(
+        self,
+        callback: Union[
+            Callable[[List[zmq.Frame]], Any],
+            Callable[[List[bytes]], Any],
+        ],
+        copy: bool = True,
+    ) -> None:
         """Register a callback for when a message is ready to recv.
-        
+
         There can be only one callback registered at a time, so each
         call to `on_recv` replaces previously registered callbacks.
-        
+
         on_recv(None) disables recv event polling.
-        
+
         Use on_recv_stream(callback) instead, to register a callback that will receive
         both this ZMQStream and the message, instead of just the message.
-        
+
         Parameters
         ----------
-        
+
         callback : callable
             callback must take exactly one argument, which will be a
             list, as returned by socket.recv_multipart()
@@ -168,96 +207,156 @@ class ZMQStream(object):
             copy is passed directly to recv, so if copy is False,
             callback will receive Message objects. If copy is True,
             then callback will receive bytes/str objects.
-        
+
         Returns : None
         """
-        
+
         self._check_closed()
         assert callback is None or callable(callback)
-        self._recv_callback = stack_context.wrap(callback)
+        self._recv_callback = stack_context_wrap(callback)
         self._recv_copy = copy
         if callback is None:
             self._drop_io_state(zmq.POLLIN)
         else:
             self._add_io_state(zmq.POLLIN)
-    
-    def on_recv_stream(self, callback, copy=True):
+
+    @overload
+    def on_recv_stream(
+        self,
+        callback: Callable[["ZMQStream", List[bytes]], Any],
+    ) -> None:
+        ...
+
+    @overload
+    def on_recv_stream(
+        self,
+        callback: Callable[["ZMQStream", List[bytes]], Any],
+        copy: Literal[True],
+    ) -> None:
+        ...
+
+    @overload
+    def on_recv_stream(
+        self,
+        callback: Callable[["ZMQStream", List[zmq.Frame]], Any],
+        copy: Literal[False],
+    ) -> None:
+        ...
+
+    @overload
+    def on_recv_stream(
+        self,
+        callback: Union[
+            Callable[["ZMQStream", List[zmq.Frame]], Any],
+            Callable[["ZMQStream", List[bytes]], Any],
+        ],
+        copy: bool = ...,
+    ):
+        ...
+
+    def on_recv_stream(
+        self,
+        callback: Union[
+            Callable[["ZMQStream", List[zmq.Frame]], Any],
+            Callable[["ZMQStream", List[bytes]], Any],
+        ],
+        copy: bool = True,
+    ):
         """Same as on_recv, but callback will get this stream as first argument
-        
+
         callback must take exactly two arguments, as it will be called as::
-        
+
             callback(stream, msg)
-        
+
         Useful when a single callback should be used with multiple streams.
         """
         if callback is None:
             self.stop_on_recv()
         else:
-            self.on_recv(lambda msg: callback(self, msg), copy=copy)
-    
-    def on_send(self, callback):
+
+            def stream_callback(msg):
+                return callback(self, msg)
+
+            self.on_recv(stream_callback, copy=copy)
+
+    def on_send(
+        self, callback: Callable[[Sequence[Any], Optional[zmq.MessageTracker]], Any]
+    ):
         """Register a callback to be called on each send
-        
+
         There will be two arguments::
-        
+
             callback(msg, status)
-        
+
         * `msg` will be the list of sendable objects that was just sent
         * `status` will be the return result of socket.send_multipart(msg) -
           MessageTracker or None.
-        
+
         Non-copying sends return a MessageTracker object whose
         `done` attribute will be True when the send is complete.
         This allows users to track when an object is safe to write to
         again.
-        
+
         The second argument will always be None if copy=True
         on the send.
-        
+
         Use on_send_stream(callback) to register a callback that will be passed
         this ZMQStream as the first argument, in addition to the other two.
-        
+
         on_send(None) disables recv event polling.
-        
+
         Parameters
         ----------
-        
+
         callback : callable
             callback must take exactly two arguments, which will be
             the message being sent (always a list),
             and the return result of socket.send_multipart(msg) -
             MessageTracker or None.
-            
+
             if callback is None, send callbacks are disabled.
         """
-        
+
         self._check_closed()
         assert callback is None or callable(callback)
-        self._send_callback = stack_context.wrap(callback)
-        
-    
-    def on_send_stream(self, callback):
+        self._send_callback = stack_context_wrap(callback)
+
+    def on_send_stream(
+        self,
+        callback: Callable[
+            ["ZMQStream", Sequence[Any], Optional[zmq.MessageTracker]], Any
+        ],
+    ):
         """Same as on_send, but callback will get this stream as first argument
-        
+
         Callback will be passed three arguments::
-        
+
             callback(stream, msg, status)
-        
+
         Useful when a single callback should be used with multiple streams.
         """
         if callback is None:
             self.stop_on_send()
         else:
             self.on_send(lambda msg, status: callback(self, msg, status))
-        
-        
+
     def send(self, msg, flags=0, copy=True, track=False, callback=None, **kwargs):
         """Send a message, optionally also register a new callback for sends.
         See zmq.socket.send for details.
         """
-        return self.send_multipart([msg], flags=flags, copy=copy, track=track, callback=callback, **kwargs)
+        return self.send_multipart(
+            [msg], flags=flags, copy=copy, track=track, callback=callback, **kwargs
+        )
 
-    def send_multipart(self, msg, flags=0, copy=True, track=False, callback=None, **kwargs):
+    def send_multipart(
+        self,
+        msg: Sequence[Any],
+        flags: int = 0,
+        copy: bool = True,
+        track: bool = False,
+        callback: Callable = None,
+        **kwargs: Any
+    ) -> None:
         """Send a multipart message, optionally also register a new callback for sends.
         See zmq.socket.send_multipart for details.
         """
@@ -270,40 +369,57 @@ class ZMQStream(object):
             # noop callback
             self.on_send(lambda *args: None)
         self._add_io_state(zmq.POLLOUT)
-    
-    def send_string(self, u, flags=0, encoding='utf-8', callback=None, **kwargs):
+
+    def send_string(
+        self,
+        u: str,
+        flags: int = 0,
+        encoding: str = 'utf-8',
+        callback: Optional[Callable] = None,
+        **kwargs: Any
+    ):
         """Send a unicode message with an encoding.
         See zmq.socket.send_unicode for details.
         """
-        if not isinstance(u, basestring):
+        if not isinstance(u, str):
             raise TypeError("unicode/str objects only")
         return self.send(u.encode(encoding), flags=flags, callback=callback, **kwargs)
-    
+
     send_unicode = send_string
-    
-    def send_json(self, obj, flags=0, callback=None, **kwargs):
+
+    def send_json(
+        self,
+        obj: Any,
+        flags: int = 0,
+        callback: Optional[Callable] = None,
+        **kwargs: Any
+    ):
         """Send json-serialized version of an object.
         See zmq.socket.send_json for details.
         """
-        if jsonapi is None:
-            raise ImportError('jsonlib{1,2}, json or simplejson library is required.')
-        else:
-            msg = jsonapi.dumps(obj)
-            return self.send(msg, flags=flags, callback=callback, **kwargs)
+        msg = jsonapi.dumps(obj)
+        return self.send(msg, flags=flags, callback=callback, **kwargs)
 
-    def send_pyobj(self, obj, flags=0, protocol=-1, callback=None, **kwargs):
+    def send_pyobj(
+        self,
+        obj: Any,
+        flags: int = 0,
+        protocol: int = -1,
+        callback: Optional[Callable] = None,
+        **kwargs: Any
+    ):
         """Send a Python object as a message using pickle to serialize.
 
         See zmq.socket.send_json for details.
         """
         msg = pickle.dumps(obj, protocol)
         return self.send(msg, flags, callback=callback, **kwargs)
-    
+
     def _finish_flush(self):
         """callback for unsetting _flushed flag."""
         self._flushed = False
-    
-    def flush(self, flag=zmq.POLLIN|zmq.POLLOUT, limit=None):
+
+    def flush(self, flag: int = zmq.POLLIN | zmq.POLLOUT, limit: Optional[int] = None):
         """Flush pending messages.
 
         This method safely handles all pending incoming and/or outgoing messages,
@@ -341,10 +457,12 @@ class ZMQStream(object):
         self._flushed = False
         # initialize counters
         count = 0
+
         def update_flag():
             """Update the poll flag, to prevent registering POLLOUT events
             if we don't have pending sends."""
             return flag & zmq.POLLIN | (self.sending() and flag & zmq.POLLOUT)
+
         flag = update_flag()
         if not flag:
             # nothing to do
@@ -352,8 +470,8 @@ class ZMQStream(object):
         self.poller.register(self.socket, flag)
         events = self.poller.poll(0)
         while events and (not limit or count < limit):
-            s,event = events[0]
-            if event & zmq.POLLIN: # receiving
+            s, event = events[0]
+            if event & zmq.POLLIN:  # receiving
                 self._handle_recv()
                 count += 1
                 if self.socket is None:
@@ -365,18 +483,18 @@ class ZMQStream(object):
                 if self.socket is None:
                     # break if socket was closed during callback
                     break
-            
+
             flag = update_flag()
             if flag:
                 self.poller.register(self.socket, flag)
                 events = self.poller.poll(0)
             else:
                 events = []
-        if count: # only bypass loop if we actually flushed something
+        if count:  # only bypass loop if we actually flushed something
             # skip send/recv callbacks this iteration
             self._flushed = True
             # reregister them at the end of the loop
-            if not already_flushed: # don't need to do it again
+            if not already_flushed:  # don't need to do it again
                 self.io_loop.add_callback(self._finish_flush)
         elif already_flushed:
             self._flushed = True
@@ -385,17 +503,17 @@ class ZMQStream(object):
         self._rebuild_io_state()
         return count
 
-    def set_close_callback(self, callback):
+    def set_close_callback(self, callback: Optional[Callable]):
         """Call the given callback when the stream is closed."""
-        self._close_callback = stack_context.wrap(callback)
+        self._close_callback = stack_context_wrap(callback)
 
-    def close(self, linger=None):
+    def close(self, linger: Optional[int] = None) -> None:
         """Close this stream."""
         if self.socket is not None:
             if self.socket.closed:
                 # fallback on raw fd for closed sockets
                 # hopefully this happened promptly after close,
-                # otherwise somebody else may have
+                # otherwise somebody else may have the FD
                 warnings.warn(
                     "Unregistering FD %s after closing socket. "
                     "This could result in unregistering handlers for the wrong socket. "
@@ -406,21 +524,28 @@ class ZMQStream(object):
                 self.io_loop.remove_handler(self._fd)
             else:
                 self.io_loop.remove_handler(self.socket)
-            self.socket.close(linger)
-            self.socket = None
+                self.socket.close(linger)
+            self.socket = None  # type: ignore
             if self._close_callback:
                 self._run_callback(self._close_callback)
 
-    def receiving(self):
+    def receiving(self) -> bool:
         """Returns True if we are currently receiving from the stream."""
         return self._recv_callback is not None
 
-    def sending(self):
+    def sending(self) -> bool:
         """Returns True if we are currently sending to the stream."""
         return not self._send_queue.empty()
 
-    def closed(self):
-        return self.socket is None
+    def closed(self) -> bool:
+        if self.socket is None:
+            return True
+        if self.socket.closed:
+            # underlying socket has been closed, but not by us!
+            # trigger our cleanup
+            self.close()
+            return True
+        return False
 
     def _run_callback(self, callback, *args, **kwargs):
         """Wrap running callbacks in try/except to allow us to
@@ -428,11 +553,9 @@ class ZMQStream(object):
         try:
             # Use a NullContext to ensure that all StackContexts are run
             # inside our blanket exception handler rather than outside.
-            with stack_context.NullContext():
-                callback(*args, **kwargs)
-        except:
-            gen_log.error("Uncaught exception in ZMQStream callback",
-                          exc_info=True)
+            callback(*args, **kwargs)
+        except Exception:
+            gen_log.error("Uncaught exception in ZMQStream callback", exc_info=True)
             # Re-raise the exception so that IOLoop.handle_callback_exception
             # can see it and log the error
             raise
@@ -441,9 +564,13 @@ class ZMQStream(object):
         """This method is the actual handler for IOLoop, that gets called whenever
         an event on my socket is posted. It dispatches to _handle_recv, etc."""
         if not self.socket:
-            gen_log.warning("Got events for closed stream %s", fd)
+            gen_log.warning("Got events for closed stream %s", self)
+            return
+        try:
+            zmq_events = self.socket.EVENTS
+        except zmq.ContextTerminated:
+            gen_log.warning("Got events for stream %s after terminating context", self)
             return
-        zmq_events = self.socket.EVENTS
         try:
             # dispatch events:
             if zmq_events & zmq.POLLIN and self.receiving():
@@ -458,8 +585,7 @@ class ZMQStream(object):
             # rebuild the poll state
             self._rebuild_io_state()
         except Exception:
-            gen_log.error("Uncaught exception in zmqstream callback",
-                          exc_info=True)
+            gen_log.error("Uncaught exception in zmqstream callback", exc_info=True)
             raise
 
     def _handle_recv(self):
@@ -478,7 +604,6 @@ class ZMQStream(object):
             if self._recv_callback:
                 callback = self._recv_callback
                 self._run_callback(callback, msg)
-        
 
     def _handle_send(self):
         """Handle a send event."""
@@ -487,7 +612,7 @@ class ZMQStream(object):
         if not self.sending():
             gen_log.error("Shouldn't have handled a send event")
             return
-        
+
         msg, kwargs = self._send_queue.get()
         try:
             status = self.socket.send_multipart(msg, **kwargs)
@@ -497,11 +622,11 @@ class ZMQStream(object):
         if self._send_callback:
             callback = self._send_callback
             self._run_callback(callback, msg, status)
-    
+
     def _check_closed(self):
         if not self.socket:
-            raise IOError("Stream is closed")
-    
+            raise OSError("Stream is closed")
+
     def _rebuild_io_state(self):
         """rebuild io state based on self.sending() and receiving()"""
         if self.socket is None:
@@ -533,10 +658,8 @@ class ZMQStream(object):
         if state & self.socket.events:
             # events still exist that haven't been processed
             # explicitly schedule handling to avoid missing events due to edge-triggered FDs
-            self.io_loop.add_callback(lambda : self._handle_events(self.socket, 0))
+            self.io_loop.add_callback(lambda: self._handle_events(self.socket, 0))
 
     def _init_io_state(self):
         """initialize the ioloop event handler"""
-        with stack_context.NullContext():
-            self.io_loop.add_handler(self.socket, self._handle_events, self.io_loop.READ)
-
+        self.io_loop.add_handler(self.socket, self._handle_events, self.io_loop.READ)
